#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>


/*
The new improved fftease.h
*/

#define getbytes t_getbytes
#define freebytes t_freebytes
#define resizebytes t_resizebytes

#define FFTEASE_ANNOUNCEMENT "- a member of FFTease 2.52"

#ifndef PI
#define PI 3.141592653589793115997963468544185161590576171875
#endif

#ifndef TWOPI
#define TWOPI 6.28318530717958623199592693708837032318115234375
#endif

#define MAX_N (16384)
#define MAX_N2 (MAX_N/2)
#define MAX_Nw (MAX_N * 4)

/* S is a spectrum in rfft format, i.e., it contains N real values
 arranged as real followed by imaginary values, except for first
 two values, which are real parts of 0 and Nyquist frequencies;
 convert first changes these into N/2+1 PAIRS of magnitude and
 phase values to be stored in output array C; the phases are then
 unwrapped and successive phase differences are used to compute
 estimates of the instantaneous frequencies for each phase vocoder
 analysis channel; decimation rate D and sampling rate R are used
 to render these frequency values directly in Hz. */

void convert(float *S, float *C, int N2, float *lastphase, float fundamental, float factor )
{
    float 	phase,
    phasediff;
    int 		real,
    imag,
    amp,
    freq;
    float 	a,
    b;
    int 		i;
    
    /*  float myTWOPI, myPI; */
    /*  double sin(), cos(), atan(), hypot();*/
    
    /*  myTWOPI = 8.*atan(1.);
     myPI = 4.*atan(1.); */
    
    
    for ( i = 0; i <= N2; i++ ) {
        imag = freq = ( real = amp = i<<1 ) + 1;
        a = ( i == N2 ? S[1] : S[real] );
        b = ( i == 0 || i == N2 ? 0. : S[imag] );
        
        C[amp] = hypot( a, b );
        if ( C[amp] == 0. )
            phasediff = 0.;
        else {
            phasediff = ( phase = -atan2( b, a ) ) - lastphase[i];
            lastphase[i] = phase;
            
            while ( phasediff > PI )
                phasediff -= TWOPI;
            while ( phasediff < -PI )
                phasediff += TWOPI;
        }
        C[freq] = phasediff*factor + i*fundamental;
    }
}
;
void unconvert( float *C, float *S, int N2, float *lastphase, float fundamental,  float factor );
void rfft( float *x, int N, int forward );
void cfft( float *x, int NC, int forward );
void bitreverse( float *x, int N );
/*
 * multiply current input I by window W (both of length Nw);
 * using modulus arithmetic, fold and rotate windowed input
 * into output array O of (FFT) length N according to current
 * input time n
 */
void fold( float *I, float *W, int Nw, float *O, int N, int n )

{
    
    int i;
    
    for ( i = 0; i < N; i++ )
        O[i] = 0.;
    
    while ( n < 0 )
      	n += N;
    n %= N;
    for ( i = 0; i < Nw; i++ ) {
        O[n] += I[i]*W[i];
      	if ( ++n == N )
            n = 0;
    }
}

;
void init_rdft(int n, int *ip, float *w)
{
    
    int	nw,
	nc;
    
    void	makewt(int nw, int *ip, float *w);
    void	makect(int nc, int *ip, float *c);
    
    nw = n >> 2;
    makewt(nw, ip, w);
    
    nc = n >> 2;
    makect(nc, ip, w + nw);
    
    return;
}
;

void rdft(int n, int isgn, float *a, int *ip, float *w)
{
    
    int		j,
    nw,
    nc;
    
    float		xi;
    
    void		bitrv2(int n, int *ip, float *a),
    cftsub(int n, float *a, float *w),
    rftsub(int n, float *a, int nc, float *c);
    
    
    nw = ip[0];
    nc = ip[1];
    
    if (isgn < 0) {
        a[1] = 0.5 * (a[1] - a[0]);
        a[0] += a[1];
        
        for (j = 3; j <= n - 1; j += 2) {
            a[j] = -a[j];
        }
        
        if (n > 4) {
            rftsub(n, a, nc, w + nw);
            bitrv2(n, ip + 2, a);
        }
        
        cftsub(n, a, w);
        
        for (j = 1; j <= n - 1; j += 2) {
            a[j] = -a[j];
        }
    }
    
    else {
        
        if (n > 4) {
            bitrv2(n, ip + 2, a);
        }
        
        cftsub(n, a, w);
        
        if (n > 4) {
            rftsub(n, a, nc, w + nw);
        }
        
        xi = a[0] - a[1];
        a[0] += a[1];
        a[1] = xi;
    }
}

;
void bitrv2(int n, int *ip, float *a)
{
    int j, j1, k, k1, l, m, m2;
    float xr, xi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    
    while ((m << 2) < l) {
        l >>= 1;
        for (j = 0; j <= m - 1; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    
    if ((m << 2) > l) {
        
        for (k = 1; k <= m - 1; k++) {
            
            for (j = 0; j <= k - 1; j++) {
                j1 = (j << 1) + ip[k];
                k1 = (k << 1) + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                a[j1] = a[k1];
                a[j1 + 1] = a[k1 + 1];
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
        }
    }
    
    else {
        m2 = m << 1;
        
        for (k = 1; k <= m - 1; k++) {
            
            for (j = 0; j <= k - 1; j++) {
                j1 = (j << 1) + ip[k];
                k1 = (k << 1) + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                a[j1] = a[k1];
                a[j1 + 1] = a[k1 + 1];
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = a[j1 + 1];
                a[j1] = a[k1];
                a[j1 + 1] = a[k1 + 1];
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
        }
    }
}
;


void cftsub(int n, float *a, float *w)
{
    int j, j1, j2, j3, k, k1, ks, l, m;
    float wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    float x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    
    while ((l << 1) < n) {
        m = l << 2;
        
        for (j = 0; j <= l - 2; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i - x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i + x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i - x3r;
        }
        
        if (m < n) {
            wk1r = w[2];
            
            for (j = m; j <= l + m - 2; j += 2) {
                j1 = j + l;
                j2 = j1 + l;
                j3 = j2 + l;
                x0r = a[j] + a[j1];
                x0i = a[j + 1] + a[j1 + 1];
                x1r = a[j] - a[j1];
                x1i = a[j + 1] - a[j1 + 1];
                x2r = a[j2] + a[j3];
                x2i = a[j2 + 1] + a[j3 + 1];
                x3r = a[j2] - a[j3];
                x3i = a[j2 + 1] - a[j3 + 1];
                a[j] = x0r + x2r;
                a[j + 1] = x0i + x2i;
                a[j2] = x2i - x0i;
                a[j2 + 1] = x0r - x2r;
                x0r = x1r - x3i;
                x0i = x1i + x3r;
                a[j1] = wk1r * (x0r - x0i);
                a[j1 + 1] = wk1r * (x0r + x0i);
                x0r = x3i + x1r;
                x0i = x3r - x1i;
                a[j3] = wk1r * (x0i - x0r);
                a[j3 + 1] = wk1r * (x0i + x0r);
            }
            
            k1 = 1;
            ks = -1;
            
            for (k = (m << 1); k <= n - m; k += m) {
                k1++;
                ks = -ks;
                wk1r = w[k1 << 1];
                wk1i = w[(k1 << 1) + 1];
                wk2r = ks * w[k1];
                wk2i = w[k1 + ks];
                wk3r = wk1r - 2 * wk2i * wk1i;
                wk3i = 2 * wk2i * wk1r - wk1i;
                
                for (j = k; j <= l + k - 2; j += 2) {
                    j1 = j + l;
                    j2 = j1 + l;
                    j3 = j2 + l;
                    x0r = a[j] + a[j1];
                    x0i = a[j + 1] + a[j1 + 1];
                    x1r = a[j] - a[j1];
                    x1i = a[j + 1] - a[j1 + 1];
                    x2r = a[j2] + a[j3];
                    x2i = a[j2 + 1] + a[j3 + 1];
                    x3r = a[j2] - a[j3];
                    x3i = a[j2 + 1] - a[j3 + 1];
                    a[j] = x0r + x2r;
                    a[j + 1] = x0i + x2i;
                    x0r -= x2r;
                    x0i -= x2i;
                    a[j2] = wk2r * x0r - wk2i * x0i;
                    a[j2 + 1] = wk2r * x0i + wk2i * x0r;
                    x0r = x1r - x3i;
                    x0i = x1i + x3r;
                    a[j1] = wk1r * x0r - wk1i * x0i;
                    a[j1 + 1] = wk1r * x0i + wk1i * x0r;
                    x0r = x1r + x3i;
                    x0i = x1i - x3r;
                    a[j3] = wk3r * x0r - wk3i * x0i;
                    a[j3 + 1] = wk3r * x0i + wk3i * x0r;
                }
            }
        }
        
        l = m;
    }
    
    if (l < n) {
        
        for (j = 0; j <= l - 2; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = a[j + 1] - a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] += a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}
;

void rftsub(int n, float *a, int nc, float *c)
{
    int j, k, kk, ks;
    float wkr, wki, xr, xi, yr, yi;
    
    ks = (nc << 2) / n;
    kk = 0;
    
    for (k = (n >> 1) - 2; k >= 2; k -= 2) {
        j = n - k;
        kk += ks;
        wkr = 0.5 - c[kk];
        wki = c[nc - kk];
        xr = a[k] - a[j];
        xi = a[k + 1] + a[j + 1];
        yr = wkr * xr - wki * xi;
        yi = wkr * xi + wki * xr;
        a[k] -= yr;
        a[k + 1] -= yi;
        a[j] += yr;
        a[j + 1] -= yi;
    }
}
;

void makewt(int nw, int *ip, float *w)
{
    void bitrv2(int n, int *ip, float *a);
    int nwh, j;
    float delta, x, y;
    
    ip[0] = nw;
    ip[1] = 1;
    if (nw > 2) {
        nwh = nw >> 1;
        delta = atan(1.0) / nwh;
        w[0] = 1;
        w[1] = 0;
        w[nwh] = cos(delta * nwh);
        w[nwh + 1] = w[nwh];
        for (j = 2; j <= nwh - 2; j += 2) {
            x = cos(delta * j);
            y = sin(delta * j);
            w[j] = x;
            w[j + 1] = y;
            w[nw - j] = y;
            w[nw - j + 1] = x;
        }
        bitrv2(nw, ip + 2, w);
    }
}

;
void makect(int nc, int *ip, float *c)
{
    int nch, j;
    float delta;
    
    ip[1] = nc;
    if (nc > 1) {
        nch = nc >> 1;
        delta = atan(1.0) / nch;
        c[0] = 0.5;
        c[nch] = 0.5 * cos(delta * nch);
        for (j = 1; j <= nch - 1; j++) {
            c[j] = 0.5 * cos(delta * j);
            c[nc - j] = 0.5 * sin(delta * j);
        }
    }
}
;
void leanconvert( float *S, float *C, int N2 );
void leanunconvert( float *C, float *S, int N2 );
void makewindows( float *H, float *A, float *S, int Nw, int N, int I )

{
    int i ;
    float sum ;
    
    for ( i = 0 ; i < Nw ; i++ )
        H[i] = A[i] = S[i] = 0.54 - 0.46*cos( TWOPI*i/(Nw - 1) ) ;
    
    if ( Nw > N ) {
        float x ;
        
        x = -(Nw - 1)/2. ;
        for ( i = 0 ; i < Nw ; i++, x += 1. )
            if ( x != 0. ) {
                A[i] *= N*sin( PI*x/N )/(PI*x) ;
                if ( I )
                    S[i] *= I*sin( PI*x/I )/(PI*x) ;
            }
    }
    
    for ( sum = i = 0 ; i < Nw ; i++ )
        sum += A[i] ;
    
    for ( i = 0 ; i < Nw ; i++ ) {
        float afac = 2./sum ;
        float sfac = Nw > N ? 1./afac : afac ;
        A[i] *= afac ;
        S[i] *= sfac ;
    }
    
    if ( Nw <= N && I ) {
        for ( sum = i = 0 ; i < Nw ; i += I )
            sum += S[i]*S[i] ;
        for ( sum = 1./sum, i = 0 ; i < Nw ; i++ )
            S[i] *= sum ;
    }
}
;
void makehamming( float *H, float *A, float *S, int Nw, int N, int I, int odd )

{
    int i;
    float sum ;
    
    
    
    if (odd) {
        for ( i = 0 ; i < Nw ; i++ )
            H[i] = A[i] = S[i] = sqrt(0.54 - 0.46*cos( TWOPI*i/(Nw - 1) ));
    }
	
    else {
        
        for ( i = 0 ; i < Nw ; i++ )
            H[i] = A[i] = S[i] = 0.54 - 0.46*cos( TWOPI*i/(Nw - 1) );
        
    }
 	
    if ( Nw > N ) {
        float x ;
        
        x = -(Nw - 1)/2. ;
        for ( i = 0 ; i < Nw ; i++, x += 1. )
            if ( x != 0. ) {
                A[i] *= N*sin( PI*x/N )/(PI*x) ;
                if ( I )
                    S[i] *= I*sin( PI*x/I )/(PI*x) ;
            }
    }
    for ( sum = i = 0 ; i < Nw ; i++ )
        sum += A[i] ;
    
    for ( i = 0 ; i < Nw ; i++ ) {
        float afac = 2./sum ;
        float sfac = Nw > N ? 1./afac : afac ;
        A[i] *= afac ;
        S[i] *= sfac ;
    }
    
    if ( Nw <= N && I ) {
        for ( sum = i = 0 ; i < Nw ; i += I )
            sum += S[i]*S[i] ;
        for ( sum = 1./sum, i = 0 ; i < Nw ; i++ )
            S[i] *= sum ;
    }
}

;
void makehanning( float *H, float *A, float *S, int Nw, int N, int I, int odd )
{
    int i;
    float sum ;
    
    
    if (odd) {
        for ( i = 0 ; i < Nw ; i++ )
            H[i] = A[i] = S[i] = sqrt(0.5 * (1. + cos(PI + TWOPI * i / (Nw - 1))));
    }
	
    else {
        
        for ( i = 0 ; i < Nw ; i++ )
            H[i] = A[i] = S[i] = 0.5 * (1. + cos(PI + TWOPI * i / (Nw - 1)));
        
    }
 	
    if ( Nw > N ) {
        float x ;
        
        x = -(Nw - 1)/2. ;
        for ( i = 0 ; i < Nw ; i++, x += 1. )
            if ( x != 0. ) {
                A[i] *= N*sin( PI*x/N )/(PI*x) ;
                if ( I )
                    S[i] *= I*sin( PI*x/I )/(PI*x) ;
            }
    }
    for ( sum = i = 0 ; i < Nw ; i++ )
        sum += A[i] ;
    
    for ( i = 0 ; i < Nw ; i++ ) {
        float afac = 2./sum ;
        float sfac = Nw > N ? 1./afac : afac ;
        A[i] *= afac ;
        S[i] *= sfac ;
    }
    
    if ( Nw <= N && I ) {
        for ( sum = i = 0 ; i < Nw ; i += I )
            sum += S[i]*S[i] ;
        for ( sum = 1./sum, i = 0 ; i < Nw ; i++ )
            S[i] *= sum ;
    }
}
;
void overlapadd( float *I, int N, float *W, float *O, int Nw, int n )

{
    int i ;
    while ( n < 0 )
        n += N ;
    n %= N ;
    for ( i = 0 ; i < Nw ; i++ ) {
        O[i] += I[n]*W[i] ;
        if ( ++n == N )
            n = 0 ;
    }
}
;
void bloscbank( float *S, float *O, int D, float iD, float *lf, float *la, float *index, float *tab, 
               int len, float synt, int lo, int hi )
{
    int    amp,freq,chan, i;
    
    float    a,ainc,f,finc,address;
    
    for ( chan = lo; chan < hi; chan++ ) {
        
        freq = ( amp = ( chan << 1 ) ) + 1;
        if ( S[amp] > synt ){ 
            finc = ( S[freq] - ( f = lf[chan] ) )* iD;
            ainc = ( S[amp] - ( a = la[chan] ) )* iD;
            address = index[chan];
            for ( i = 0; i < D ; i++ ) {
                O[i] += a*tab[ (int) address ];
                
                address += f;
                while ( address >= len )
                    address -= len;
                while ( address < 0 )
                    address += len;
                a += ainc;
                f += finc;
            }
            lf[chan] = S[freq];
            la[chan] = S[amp];
            index[chan] = address;
        }    
    }
}
;

float randf( float min, float max );
int randi( int min, int max );
int power_of_two(int test)
{
    int limit = 8192;
    int compare = 1;
    //  post("testing what we thing is an int:%d",test);
    do {
        if(test == compare){
            //      post("good power of 2 found!");
            return 1;
        } 
        compare *= 2;
    } while (compare <= limit);
    
    return 0;
}
// we think the above function is shadowed somewhere else

void freebytes2(void *fatso, size_t nbytes);
void *getbytes2(size_t nbytes);
void *resizebytes2(void *old, size_t oldsize, size_t newsize);
void limit_fftsize(int *N, int *Nw, char *OBJECT_NAME)
{
	if(*N > MAX_N){
		// post("%s: N set to maximum FFT size of %d",OBJECT_NAME,MAX_N); 
		*N = MAX_N;
	}
	if(*Nw > MAX_Nw){
        // 	post("%s: Nw set to maximum window size of %d",OBJECT_NAME,MAX_Nw); 
		*Nw = MAX_Nw;
	}
};

/* THE FUNCTIONS */
